/* 
 * sdui-tty.c - SD TTY User Interface
 * Originally for Macintosh.  Unix version by gildea.
 * Time-stamp: <94/11/10 17:17:53 wba>
 * Copyright (c) 1990-1994 Stephen Gildea, William B. Ackerman, and
 *   Alan Snyder
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose is hereby granted without fee, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * The authors make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 *
 * By Alan Snyder, December 1992.
 *
 * This version of the Sd UI does completing reads instead of menus.
 * Type '?' to see possibilities.
 *    At the "More" prompt, type SPC to see the next page, DEL to stop.
 * Type SPC to complete the current word.
 * Type TAB to complete as much as possible.
 * Type Control-U to clear the line.
 *
 * For use with version 30 of the Sd program.
 * Based on sdui-x11.c 1.10
 *
 * The version of this file is as shown immediately below.  This string
 * gets displayed at program startup, as the "ui" part of the complete
 * version.
 */

#define UI_VERSION_STRING "1.6"

/* See comments in sdmain.c regarding this string. */
static char *id="@(#)$He" "ader: Sd: sdui-tty.c "
   UI_VERSION_STRING
   "  wba@apollo.hp.com  10 Nov 94 $";

/* This file defines the following functions:
   uims_process_command_line
   uims_version_string
   uims_preinitialize
   uims_create_menu
   uims_postinitialize
   uims_get_command
   uims_do_comment_popup
   uims_do_outfile_popup
   uims_do_getout_popup
   uims_do_abort_popup
   uims_do_neglect_popup
   uims_do_selector_popup
   uims_do_direction_popup
   uims_do_tagger_popup
   uims_get_number_fields
   uims_do_modifier_popup
   uims_add_new_line
   uims_reduce_line_count
   uims_begin_search
   uims_begin_reconcile_history
   uims_end_reconcile_history
   uims_update_resolve_menu
   uims_terminate
   uims_database_tick_max
   uims_database_tick
   uims_database_tick_end
   uims_database_error
   uims_bad_argument
*/

#ifndef THINK_C
#define UNIX_STYLE
#endif

/* For "sprintf" and some IO stuff (fflush, printf, stdout) that we use
   during the "database tick" printing before the actual IO package is started.
   During normal operation, we don't do any IO at all in this file. */
#include <stdio.h>
/* For "strlen". */
#include <string.h>
/* For "isprint". */
#include <ctype.h>
/* For "atoi". */
#include <stdlib.h>

#include "sd.h"
#include "sdmatch.h"
#include "sdui-ttu.h"

#define DEL 0x7F
#define INPUT_TEXTLINE_SIZE 200

/*
 * The total version string looks something like
 * "1.4:db1.5:ui0.6tty"
 * We return the "0.6tty" part.
 */

Private char version_mem[12];

extern char *
uims_version_string(void)
{
    (void) sprintf(version_mem, "%stty", UI_VERSION_STRING);
    return version_mem;
}

/*
 * User Input functions
 */

/* not passed as args so refresh_input can be called from signal handler */
Private char user_input[INPUT_TEXTLINE_SIZE+1];
Private int user_input_size;
Private char *user_input_prompt;
Private char *function_key_expansion;

void
refresh_input(void)
{
    user_input[0] = '\0';
    user_input_size = 0;
    function_key_expansion = (char *) 0;
    clear_line(); /* clear the current line */
    put_line(user_input_prompt);
}

Private void
get_string_input(char prompt[], char dest[])
{
    put_line(prompt);
    get_string(dest);
}


Private int
get_char_input(void)
{
   int c;

   start_expand:

   if (function_key_expansion) {
      c = *function_key_expansion++;
      if (c) return c;
      else function_key_expansion = (char *) 0;
   }

#ifdef THINK_C			/* Unix or DOS provides its own cursor */
   putchar('_'); /* a cursor */
   c = get_char();
   putchar('\b'); putchar(' '); putchar('\b'); /* erase the cursor */
#else
   c = get_char();
#endif

   if (c == 129)
      function_key_expansion = "split phantom lines\n";
   else if (c == 130)
      function_key_expansion = "split phantom columns\n";
   else if (c == 131)
      function_key_expansion = "simple modifications\n";
   else if (c == 132)
      function_key_expansion = "<anything>";
   else if (c == 161)
      function_key_expansion = "interlocked phantom lines\n";
   else if (c == 162)
      function_key_expansion = "interlocked phantom columns\n";
   else if (c == 193)
      function_key_expansion = "phantom lines\n";
   else if (c == 194)
      function_key_expansion = "phantom columns\n";
   else if (c >= 128)
      c = ' ';

   if (function_key_expansion)
      goto start_expand;

   return c;

}
  
/* This tells how many of the last lines currently on the screen contain
   the text that the main program thinks comprise the transcript.  That is,
   if we count this far up from the bottom of the screen (well, from the
   cursor, to be precise), we will get the line that the main program thinks
   is the first line of the transcript.  This variable will often be greater
   than the main program's belief of the transcript length, because we are
   fooling around with input lines (prompts, error messages, etc.) at the
   bottom of the screen.

   The main program typically updates the transcript in a "clean" way by
   calling "uims_reduce_line_count" to erase some number of lines at the
   end of the transcript, followed by a rewrite of new material.  We compare
   what "uims_reduce_line_count" tells us against the value of this variable
   to find out how many lines to actually erase.  That way, we erase the input
   lines and produce a truly clean transcript, on devices (like VT-100's)
   capable of doing so.  If the display device can't erase lines, the user
   will see the last few lines of the transcript overlapping from one
   command to the next.  This is the appropriate behavior for a printing
   terminal or computer emulation of same. */

Private int current_text_line;

/*
 * Throw away all but the first n lines of the text output.
 * n = 0 means to erase the entire buffer.
 */

Private void
text_output_trim(int n)
{
    if (current_text_line > n)
        erase_last_n(current_text_line-n);

    current_text_line = n;
}

/*
 * end of text output stuff
 */

Private char *call_menu_prompts[NUM_CALL_LIST_KINDS];

/*
 * The main program calls this before doing anything else, so we can
 * supply additional command line arguments.
 * Note: If we are writing a call list, the program will
 * exit before doing anything else with the user interface, but this
 * must be made anyway.
 */
 
extern void
uims_process_command_line(int *argcp, char ***argvp)
{
#ifdef UNIX_STYLE
   ttu_process_command_line(argcp, argvp);
#else
   *argcp = ccommand(argvp); /* pop up window to get command line arguments */
#endif
}

/*
 * The main program calls this before any of the call menus are
 * created, that is, before any calls to uims_create_menu.
 */
 
extern void
uims_preinitialize(void)
{
}

/*
 * Create a menu containing number_of_calls[cl] items.
 * Use the "menu_names" array to create a
 * title line for the menu.  The string is in static storage.
 * 
 * This will be called once for each value in the enumeration call_list_kind.
 */

extern void uims_create_menu(call_list_kind cl, callspec_block *call_name_list[])
{
   call_menu_prompts[cl] = (char *) get_mem(50);  /* *** Too lazy to compute it. */
   matcher_setup_call_menu(cl, call_name_list);

   if (cl == call_list_any)
      /* The menu name here is "(any setup)".  That may be a sensible
         name for a menu, but it isn't helpful as a prompt.  So we
         just use a vanilla prompt. */
      call_menu_prompts[cl] = "--> ";
   else
      (void) sprintf(call_menu_prompts[cl], "(%s)--> ", menu_names[cl]);
}


/* The main program calls this after all the call menus have been created,
   after all calls to uims_create_menu.
   This performs any final initialization required by the interface package.

   It must also perform any required setup of the concept menu.  The
   concepts are not presented procedurally.  Instead, they can be found
   in the external array concept_descriptor_table+general_concept_offset.
   The number of concepts in that list is general_concept_size.  For each
   i, the field concept_descriptor_table[i].name has the text that we
   should display for the user.
*/

extern void
uims_postinitialize(void)
{
    call_menu_prompts[call_list_empty] = "--> ";   /* This prompt should never be used. */
    matcher_initialize(TRUE); /* want commands last in menu */
#if defined(UNIX_STYLE) && !defined(MSDOS)
    initialize_signal_handlers();
#endif
    resolver_is_unwieldy = TRUE;   /* Sorry about that. */
    ttu_initialize();
    current_text_line = 0;
}

Private void
pack_and_echo_character(char c)
{
   /* Really should handle error better -- ring the bell,
      but this is called inside a loop. */

   if (user_input_size < INPUT_TEXTLINE_SIZE) {
      user_input[user_input_size++] = c;
      user_input[user_input_size] = '\0';
      put_char(c);
   }
}

/* This tells how many more lines of matches (the stuff we print in response
   to a question mark) we can print before we have to say "--More--" to
   get permission from the user to continue.  If it goes negative, the
   user has given a negative reply to one of our queries, and so we don't
   print any more stuff. */
Private int match_counter;

/* This is what we reset the counter to whenever the user confirms.  That
   is, it is the number of lines we print per "screenful".  On a VT-100-like
   ("dumb") terminal, we will actually make it a screenful.  On a printing
   device or a workstation, we don't need to do the hold-screen stuff, because
   the device can handle output intelligently (on a printing device, it does
   this by letting us look at the paper that is spewing out on the floor;
   on a workstation we know that the window system provides real (infinite,
   of course) scrolling).  But on printing devices or workstations we still
   do the output in screenful-like blocks, because the user may not want
   to see an enormous amount of output. */
Private int match_lines;

Private void
start_matches(void)
{
   /*
    * Find the number of lines on the screen.
    * Used only for grouping "screenfuls" with "--more--" stuff
    * when showing long lists of choices after '?' is typed.
    * In other cases, we just assume the output mechanism can intelligently
    * handle what we send to it.
    */

    match_lines = get_lines_for_more();
    match_counter = match_lines-1; /* last line used for "--More--" prompt */
}

Private int
prompt_for_more_output(void)
{
    put_line("--More--");

    for (;;) {
        char c = get_char_input();
        clear_line();   /* Erase the "more" line; next item goes on that line. */

        switch (c) {
          case '\r':
          case '\n': match_counter = 1; /* show one more line */
                     return TRUE;  /* keep going */

          case '\b':
          case DEL:
          case 'q':
          case 'Q':  return FALSE; /* stop showing */

          case ' ':  return TRUE;  /* keep going */

          default:   put_line("Type Space to see more, Return for next line, Delete to stop:  --More--");
        }
    }
}

Private void
show_match(char *user_input_str, Const char *extension, Const match_result *mr)
{
   char temp[200];
   char c;
   Const char *p;
   char *q;
   char *z;

   if (match_counter < 0) return;  /* Showing has been turned off. */

   if (match_counter <= 0) {
       match_counter = match_lines - 1;
       if (!prompt_for_more_output()) {
           match_counter = -1;   /* Turn it off. */
           return;
       }
   }
   match_counter--;
   put_line(user_input_str);
   put_line((char *) extension);
   put_line("\n");
   current_text_line++;
}

/*
 * The searching functions indicate their results by setting the variables
 * last_match and exact_match.  Last_match indicates the last matching item
 * (if there is one or more matching item).  Exact_match is set if
 * an item matches the input string exactly.  Hopefully, if there is more than
 * one item that matches exactly, they are all equivalent.
 */

Private match_result user_match;

/* BEWARE!!  The first part of this list is keyed to the definition of
   "command_kind" in sd.h .
   There are NUM_COMMAND_KINDS items in that part.  The rest of it
   corresponds to the special commands defined in sdmatch.h .  There
   are NUM_SPECIAL_COMMANDS of those items. */

static Cstring command_list[] = {
    "exit the program",
    "undo last call",
    "abort this sequence",
    "insert a comment",
    "change output file",
    "end this sequence",
    "resolve",
    "reconcile",
    "pick random call",
    "normalize",
#ifdef NEGLECT
    "show neglected calls",
#endif
    "keep picture",
    "refresh display",
/* The following items are the special ones. */
    "simple modifications",
    "allow modifications",
    "toggle concept levels",
    "toggle active phantoms"
};

#define NUM_SPECIAL_COMMANDS 4
#define SPECIAL_COMMAND_SIMPLE_MODS 0
#define SPECIAL_COMMAND_ALLOW_MODS 1
#define SPECIAL_COMMAND_TOGGLE_CONCEPT_LEVELS 2
#define SPECIAL_COMMAND_TOGGLE_ACTIVE_PHANTOMS 3



/* result is indicated in user_match */

Private void
get_user_input(char *prompt, int which)
{
    char extended_input[200];
    char *p;
    char c;
    int matches;
    
    user_match.valid = FALSE;
    user_input_prompt = prompt;
    user_input[0] = '\0';
    user_input_size = 0;
    function_key_expansion = (char *) 0;
    put_line(user_input_prompt);

    for (;;) {
        /* At this point we always have the concatenation of "user_input_prompt"
           and "user_input" displayed on the current line. */

        c = get_char_input();

        if ((c == '\b') || (c == DEL)) {
            if (user_input_size > 0) {
                user_input_size--;
                user_input[user_input_size] = '\0';
                function_key_expansion = (char *) 0;
                rubout(); /* Update the display. */
            }
            continue;
        }
        else if (c == '?') {
            put_line ("\n");
            current_text_line++;
            start_matches();
            match_user_input(user_input, which, (match_result *) 0, command_list, NUM_COMMAND_KINDS+NUM_SPECIAL_COMMANDS, (char *) 0, show_match, FALSE);
            put_line ("\n");     /* Write a blank line. */
            current_text_line++;
            put_line(user_input_prompt);   /* Redisplay the current line. */
            put_line(user_input);
            continue;
        }

        if (c == ' ') {
            /* extend only to one space, inclusive */
            matches = match_user_input(user_input, which, &user_match, command_list, NUM_COMMAND_KINDS+NUM_SPECIAL_COMMANDS, extended_input, (show_function) 0, FALSE);
            p = extended_input;
            if (*p) {
                while (*p) {
                    if (*p != ' ')
                        pack_and_echo_character(*p++);
                    else
                        goto foobar;
                }
                continue;   /* Do *not* pack the character. */

                foobar: ;
                pack_and_echo_character(c);
            }
            else if (user_match.space_ok && matches > 1)
                pack_and_echo_character(c);
            else if (diagnostic_mode)
                goto diagnostic_error;
            else
                bell();
        }
        else if ((c == '\n') || (c == '\r')) {
            matches = match_user_input(user_input, which, &user_match, command_list, NUM_COMMAND_KINDS+NUM_SPECIAL_COMMANDS, extended_input, (show_function) 0, FALSE);
            if (matches == 1 || matches - user_match.yielding_matches == 1 || user_match.exact) {
                p = extended_input;
                while (*p)
                    pack_and_echo_character(*p++);

                put_line("\n");
                /* Include the input line in our count, so we will erase it
                   if we are trying to make the VT-100 screen look nice. */
                current_text_line++;
                return;
            }

            if (diagnostic_mode)
                goto diagnostic_error;

            /* Tell how bad it is, then redisplay current line. */
	   if (matches > 0) {
                char tempstuff[200];

                (void) sprintf(tempstuff, "  (%d matches, type ? for list)\n", matches);
                put_line(tempstuff);
            }
            else {
                put_line("  (no matches)\n");
            }

            put_line(user_input_prompt);
            put_line(user_input);
            current_text_line++;   /* Count that line for erasure. */
        }
        else if (c == '\t' || c == '\033') {
            matches = match_user_input(user_input, which, &user_match, command_list, NUM_COMMAND_KINDS+NUM_SPECIAL_COMMANDS, extended_input, (show_function) 0, FALSE);
            p = extended_input;
            if (*p) {
                while (*p)
                    pack_and_echo_character(*p++);
            }
            else if (diagnostic_mode)
                goto diagnostic_error;
            else
                bell();
        }
        else if (c == ('U'&'\037')) { /* C-u: kill line */
            user_input[0] = '\0';
            user_input_size = 0;
            function_key_expansion = (char *) 0;
            clear_line();           /* Clear the current line */
            put_line(user_input_prompt);    /* Redisplay the prompt. */
        }
        else if (isprint(c))
            pack_and_echo_character(c);
        else if (diagnostic_mode)
            goto diagnostic_error;
        else
            bell();
    }

    diagnostic_error:

    uims_terminate();
    (void) fputs("\nParsing error during diagnostic.\n", stdout);
    (void) fputs("\nParsing error during diagnostic.\n", stderr);
    final_exit(1);
}

Private call_list_kind current_call_menu;

static char *banner_prompts[] = {
    (char *) 0,
    "[AP] ",
    "[all concepts] ",
    "[all concepts,AP] ",
    "[simple modifications] ",
    "[AP,simple modifications] ",
    "[all concepts,simple modifications] ",
    "[all concepts,AP,simple modifications] ",
    "[all modifications] ",
    "[AP,all modifications] ",
    "[all concepts,all modifications] ",
    "[all concepts,AP,all modifications] "};

extern uims_reply
uims_get_command(mode_kind mode, call_list_kind *call_menu)
{
    if (mode == mode_startup) {
         get_user_input("Enter startup command> ", (int) match_startup_commands);
         uims_menu_index = user_match.index;
    }
    else if (mode == mode_resolve) {
         get_user_input("Enter resolve command> ", (int) match_resolve_commands);
         uims_menu_index = user_match.index;
    }
    else {
        char prompt_buffer[200];
        char *prompt_ptr;
        int banner_mode;

        check_menu:

        if (allowing_modifications)
            *call_menu = call_list_any;

        current_call_menu = *call_menu;
        prompt_ptr = prompt_buffer;

        /* Put any necessary special things into the prompt. */

        banner_mode = (allowing_modifications << 2) |
                      (allowing_all_concepts ? 2 : 0) |
                      (using_active_phantoms ? 1 : 0);

        if (banner_mode != 0)
            (void) sprintf(prompt_ptr, "%s%s", banner_prompts[banner_mode], call_menu_prompts[current_call_menu]);
        else
            prompt_ptr = call_menu_prompts[current_call_menu];

        get_user_input(prompt_ptr, (int) current_call_menu);

        uims_menu_index = user_match.index;
        uims_menu_cross = user_match.cross;
        uims_menu_magic = user_match.magic;
        uims_menu_intlk = user_match.interlocked;
        uims_menu_left = user_match.left;

        if (user_match.kind == ui_command_select && uims_menu_index >= NUM_COMMAND_KINDS) {
            if (uims_menu_index == NUM_COMMAND_KINDS + SPECIAL_COMMAND_SIMPLE_MODS) {
                /* Increment "allowing_modifications" up to a maximum of 2. */
                if (allowing_modifications != 2) allowing_modifications++;
            }
            else if (uims_menu_index == NUM_COMMAND_KINDS + SPECIAL_COMMAND_ALLOW_MODS) {
                allowing_modifications = 2;
            }
            else if (uims_menu_index == NUM_COMMAND_KINDS + SPECIAL_COMMAND_TOGGLE_CONCEPT_LEVELS) {
                allowing_all_concepts = !allowing_all_concepts;
            }
            else {   /* Must be SPECIAL_COMMAND_TOGGLE_ACTIVE_PHANTOMS. */
                using_active_phantoms = !using_active_phantoms;
            }

            goto check_menu;
        }
    }

    return user_match.kind;
}
                                                              

Private int
get_popup_string(char prompt[], char dest[])
{
    char buffer[200];

    (void) sprintf(buffer, "%s: ", prompt);
    get_string_input(buffer, dest);
    return POPUP_ACCEPT_WITH_STRING;
}

extern int
uims_do_comment_popup(char dest[])
{
    return get_popup_string("Enter comment", dest);
}

extern int
uims_do_outfile_popup(char dest[])
{
    char buffer[200];
    (void) sprintf(buffer, "Sequence output file is \"%s\".\n", outfile_string);

    put_line(buffer);
    current_text_line++;
    return get_popup_string("Enter new file name", dest);
}

extern int
uims_do_getout_popup(char dest[])
{
    put_line("Specify text label for sequence.\n");
    current_text_line++;
    return get_popup_string("Enter label", dest);
}

#ifdef NEGLECT
extern int
uims_do_neglect_popup(char dest[])
{
    put_line("Specify integer percentage of neglected calls.\n");
    current_text_line++;
    return get_popup_string("Enter percentage", dest);
}
#endif

Private int
confirm(char *question)
{
    char c;

    for (;;) {
        put_line(question);
        c = get_char_input();
        if ((c=='n') || (c=='N')) {
            put_line("no\n");
            current_text_line++;
            return POPUP_DECLINE;
        }
        if ((c=='y') || (c=='Y')) {
            put_line("yes\n");
            current_text_line++;
            return POPUP_ACCEPT;
        }
        put_char(c);

        if (diagnostic_mode) {
            uims_terminate();
            (void) fputs("\nParsing error during diagnostic.\n", stdout);
            (void) fputs("\nParsing error during diagnostic.\n", stderr);
            final_exit(1);
        }

        put_line("\n");
        put_line("Answer y or n\n");
        current_text_line += 2;
        bell();
    }
}

extern int
uims_do_abort_popup(void)
{
    put_line("The current sequence will be aborted.\n");
    current_text_line++;
    return confirm("Do you really want to abort it? ");
}

extern int uims_do_modifier_popup(Cstring callname, modify_popup_kind kind)
{
    char *line_format = "Internal error: unknown modifier kind.\n";
    char tempstuff[200];
    
    switch (kind) {
        case modify_popup_any:
            line_format = "The \"%s\" can be replaced.\n";
            break;
        case modify_popup_only_tag:
            line_format = "The \"%s\" can be replaced with a tagging call.\n";
            break;
        case modify_popup_only_scoot:
            line_format = "The \"%s\" can be replaced with a scoot/tag (chain thru) (and scatter).\n";
            break;
    }

    sprintf(tempstuff, line_format, callname);
    put_line(tempstuff);
    current_text_line++;
    return confirm("Do you want to replace it? ");
}

/*
 * UIMS_BEGIN_SEARCH is called at the beginning of each search mode
 * command (resolve, reconcile, nice setup, pick random call).
 */

extern void
uims_begin_search(search_kind goal)
{
}

/*
 * UIMS_BEGIN_RECONCILE_HISTORY is called at the beginning of a reconcile,
 * after UIMS_BEGIN_SEARCH,
 * and whenever the current reconcile point changes.  CURRENTPOINT is the
 * current reconcile point and MAXPOINT is the maximum possible reconcile
 * point.  This call is followed by calls to UIMS_REDUCE_LINE_COUNT
 * and UIMS_ADD_NEW_LINE that
 * display the current sequence with the reconcile point indicated.  These
 * calls are followed by a call to UIMS_END_RECONCILE_HISTORY.
 * Return TRUE to cause sd to forget its cached history output; do this
 * if the reconcile history is written to a separate window.
 */

extern int
uims_begin_reconcile_history(int currentpoint, int maxpoint)
{
    return FALSE;
}

/*
 * Return TRUE to cause sd to forget its cached history output.
 */

extern int
uims_end_reconcile_history(void)
{
    put_line("\n");
    return FALSE;
}

extern void
uims_update_resolve_menu(search_kind goal, int cur, int max, resolver_display_state state)
{
    char title[MAX_TEXT_LINE_LENGTH];

    create_resolve_menu_title(goal, cur, max, state, title);
    put_line(title);
    put_line("\n");
    current_text_line++;
}

extern int
uims_do_selector_popup(void)
{
    int n;

    if (user_match.valid && (user_match.who > selector_uninitialized)) {
        n = (int) user_match.who;
        user_match.who = selector_uninitialized;
        return n;
    }
    else {
        /* We skip the zeroth selector, which is selector_uninitialized. */
        get_user_input("Enter who> ", (int) match_selectors);
        return user_match.index+1;
    }
}    

extern int
uims_do_direction_popup(void)
{
    int n;

    if (user_match.valid && (user_match.where > direction_uninitialized)) {
        n = (int) user_match.where;
        user_match.where = direction_uninitialized;
        return n;
    }
    else {
        /* We skip the zeroth direction, which is direction_uninitialized. */
        get_user_input("Enter direction> ", (int) match_directions);
        return user_match.index+1;
    }
}    

extern int
uims_do_tagger_popup(void)
{
   if (user_match.valid && (user_match.tagger > 0)) {
     uint32 n;
     int j = 0;

      while ((user_match.tagger & 0xFF000000) == 0) {
         user_match.tagger <<= 8;
         j++;
      }

      n = user_match.tagger >> 24;
      user_match.tagger &= 0x00FFFFFF;
      while (j-- != 0) user_match.tagger >>= 8;   /* Shift it back. */
      return n;
   }
   else {
      get_user_input("Enter tagger> ", (int) match_taggers);
      return user_match.index+1;
   }
}    


extern uint32 uims_get_number_fields(int nnumbers)
{
   int i;
   char buffer[200];
   unsigned int number_list = 0;

   for (i=0 ; i<nnumbers ; i++) {
      unsigned int this_num;

      if (user_match.valid && (user_match.howmanynumbers >= 1)) {
         this_num = user_match.number_fields & 0xF;
         user_match.number_fields >>= 4;
         user_match.howmanynumbers--;
      }
      else {
         get_string_input("How many? ", buffer);
         this_num = atoi(buffer);
      }

      if (this_num == 0 || this_num > 8) return 0;    /* User gave bad answer. */
      number_list |= (this_num << (i*4));
   }

   return number_list;
}


/*
 * add a line to the text output area.
 * the_line does not have the trailing Newline in it and
 * is volatile, so we must copy it if we need it to stay around.
 */

extern void
uims_add_new_line(char the_line[])
{
    current_text_line++;
    put_line(the_line);
    put_line("\n");
}

/*
 * Throw away all but the first n lines of the text output.
 * n = 0 means to erase the entire buffer.
 */

extern void
uims_reduce_line_count(int n)
{
    text_output_trim(n);
}

extern void
uims_terminate(void)
{
    ttu_terminate();
}

/*
 * The following two functions allow the UI to put up a progress
 * indicator while the call database is being read (and processed).
 *
 * uims_database_tick_max is called before reading the database
 * with the number of ticks that will be sent.
 * uims_database_tick is called repeatedly with the number of new
 * ticks to add.
 */

Private int db_tick_max;
Private int db_tick_cur;	/* goes from 0 to db_tick_max */

#define TICK_STEPS 52
Private int tick_displayed;	/* goes from 0 to TICK_STEPS */

extern void
uims_database_tick_max(int n)
{
   db_tick_max = n;
   db_tick_cur = 0;
   printf("Sd: reading database...");
   fflush(stdout);
   tick_displayed = 0;
}

extern void
uims_database_tick(int n)
{
    int tick_new;

    db_tick_cur += n;
    tick_new = TICK_STEPS*db_tick_cur/db_tick_max;
    while (tick_displayed < tick_new) {
	printf(".");
	tick_displayed++;
    }
    fflush(stdout);
}

extern void
uims_database_tick_end(void)
{
    printf("done\n");
}

extern void
uims_database_error(char *message, char *call_name)
{
   print_line(message);
   if (call_name) {
      print_line("  While reading this call from the database:");
      print_line(call_name);
   }
}

extern void
uims_bad_argument(char *s1, char *s2, char *s3)
{
   if (s1) print_line(s1);
   if (s2) print_line(s2);
   if (s3) print_line(s3);
   print_line("Use the -help flag for help.");
   exit_program(1);
}
